use std::{sync::Arc, time::Duration};

use anyhow::Result;
use dbus::{
    nonblock::{Proxy, SyncConnection},
    Path,
};
use num_traits::FromPrimitive;

use nm_reactor_dbus_codegen::network_manager_device::{
    OrgFreedesktopNetworkManagerDevice, OrgFreedesktopNetworkManagerDeviceStateChanged,
};

use super::{
    active_connection::ActiveConnectionWrapper, device_state::DeviceState,
    signal::SignalStreamWrapper,
};

pub struct DeviceWrapper<'a> {
    conn: Arc<SyncConnection>,
    proxy: Box<dyn OrgFreedesktopNetworkManagerDevice + Send + Sync + 'a>,
    path: Path<'a>,
}

impl<'a> DeviceWrapper<'a> {
    pub async fn from_path<P>(conn: Arc<SyncConnection>, path: P) -> DeviceWrapper<'a>
    where
        P: Into<dbus::Path<'a>>,
    {
        let path = path.into();

        let proxy: Proxy<'a, Arc<SyncConnection>> = Proxy::new(
            "org.freedesktop.NetworkManager",
            path.clone(),
            Duration::from_millis(5000),
            conn.clone(),
        );

        DeviceWrapper {
            conn,
            proxy: Box::new(proxy),
            path,
        }
    }

    pub fn get_path(&self) -> Path<'a> {
        self.path.clone()
    }

    pub async fn get_active_connection(&self) -> Result<Option<ActiveConnectionWrapper<'static>>> {
        let path = self.proxy.active_connection().await?;

        if path == Path::from("/") {
            return Ok(None);
        }

        let connection = ActiveConnectionWrapper::from_path(self.conn.clone(), path).await;

        Ok(Some(connection))
    }

    pub async fn state_changed_signal_stream(
        &self,
    ) -> anyhow::Result<SignalStreamWrapper<OrgFreedesktopNetworkManagerDeviceStateChanged>> {
        SignalStreamWrapper::from_match_rule(&self.conn, None, Some(self.path.clone())).await
    }

    pub async fn get_state(&self) -> Result<Option<DeviceState>> {
        let result = self.proxy.state().await?;

        let result = FromPrimitive::from_u32(result);

        Ok(result)
    }
}
