use std::sync::Arc;

use dbus::{
    arg::ReadAll,
    message::{MatchRule, SignalArgs},
    nonblock::{MsgMatch, SyncConnection},
    strings::BusName,
    Message, MessageType, Path,
};
use futures_channel::mpsc::UnboundedReceiver;
use tokio_stream::StreamExt;

pub struct SignalStreamWrapper<T> {
    conn: Arc<SyncConnection>,
    msg_match: MsgMatch,
    stream: UnboundedReceiver<(Message, T)>,
}

impl<T> SignalStreamWrapper<T> {
    pub fn from_stream(
        conn: &Arc<SyncConnection>,
        msg_match: MsgMatch,
        stream: UnboundedReceiver<(Message, T)>,
    ) -> SignalStreamWrapper<T> {
        SignalStreamWrapper {
            conn: conn.clone(),
            msg_match,
            stream,
        }
    }

    pub async fn from_match_rule(
        conn: &Arc<SyncConnection>,
        sender: Option<BusName<'_>>,
        path: Option<Path<'_>>,
    ) -> anyhow::Result<SignalStreamWrapper<T>>
    where
        T: ReadAll + Send + SignalArgs + 'static,
    {
        let mut match_rule = MatchRule::default();

        match_rule.sender = sender.map(|s| s.into_static());
        match_rule.path = path.map(|p| p.into_static());
        match_rule.msg_type = Some(MessageType::Signal);
        match_rule.interface = Some(T::INTERFACE.into());
        match_rule.member = Some(T::NAME.into());

        let (msg_match, stream) = conn.add_match(match_rule).await?.stream::<T>();

        Ok(Self::from_stream(conn, msg_match, stream))
    }

    pub async fn next(&mut self) -> Option<(Message, T)> {
        self.stream.next().await
    }

    pub async fn dispose(&self) -> anyhow::Result<()> {
        self.conn.remove_match(self.msg_match.token()).await?;

        Ok(())
    }
}
