use std::sync::Arc;

use anyhow::Result;
use dbus::{nonblock::SyncConnection, Path};
use num_traits::FromPrimitive;
use tokio::{
    sync::{broadcast, mpsc},
    task::JoinHandle,
};

use crate::{
    dbus_wrappers::{device::DeviceWrapper, device_state::DeviceState},
    event::Event,
};

pub async fn watch_device(
    conn: &Arc<SyncConnection>,
    mut stop_signal_rx: broadcast::Receiver<()>,
    event_tx: mpsc::Sender<Event>,
    device_path: Path<'_>,
) -> Result<JoinHandle<Result<()>>> {
    log::debug!("Setting up signal handlers for {}", device_path);

    let conn_for_task = conn.clone();
    let device_path_for_task = device_path.into_static();

    let handler_handle: JoinHandle<Result<()>> = tokio::spawn(async move {
        let device = DeviceWrapper::from_path(conn_for_task.clone(), device_path_for_task).await;

        let mut state_changed_signal = device.state_changed_signal_stream().await?;

        loop {
            tokio::select! {
                Some((msg, signal)) = state_changed_signal.next() => {
                    log::debug!("Got StateChanged signal for {}: {:?}", device.get_path(), &msg);

                    match FromPrimitive::from_u32(signal.new_state) {
                        Some(DeviceState::Activated) => {
                            event_tx.send(Event::DeviceActivated {
                                device_path: device.get_path(),
                            }).await?;
                        }
                        Some(DeviceState::Deactivating) => {
                            event_tx.send(Event::DeviceDeactivating {
                                device_path: device.get_path(),
                            }).await?;
                        }
                        Some(DeviceState::Disconnected) => {
                            event_tx.send(Event::DeviceDisconnected {
                                device_path: device.get_path(),
                            }).await?;
                        }
                        _ => {}
                    }
                },
                _ = stop_signal_rx.recv() => {
                    log::debug!("Stoping device watcher for {}", device.get_path());

                    break;
                }
            }
        }

        log::debug!("Removing signal handlers for {}", device.get_path());

        state_changed_signal.dispose().await?;

        Ok(())
    });

    Ok(handler_handle)
}
