use std::{collections::HashSet, sync::Arc};

use anyhow::Result;
use dbus::nonblock::SyncConnection;
use serde_derive::{Deserialize, Serialize};
use tokio::{sync::mpsc, task::JoinHandle};

use crate::{
    config::Config,
    dbus_wrappers::{connectivity_state::ConnectivityState, state::State},
    identifier::DeviceIdentifier,
};

#[derive(Debug)]
pub enum Event {
    ConnectivityStateChanged {
        connectivity_state: ConnectivityState,
    },
    DaemonStarted,
    DaemonStopping,
    DeviceAdded {
        device_path: dbus::Path<'static>,
    },
    DeviceRemoved {
        device_path: dbus::Path<'static>,
    },
    DeviceActivated {
        device_path: dbus::Path<'static>,
    },
    DeviceDeactivating {
        device_path: dbus::Path<'static>,
    },
    DeviceDisconnected {
        device_path: dbus::Path<'static>,
    },
    StateChanged {
        state: State,
    },
}

pub async fn handle_events(
    conn: Arc<SyncConnection>,
    config: Arc<Config>,
    mut event_rx: mpsc::Receiver<Event>,
) -> Result<JoinHandle<Result<()>>> {
    log::info!("Starting event handler task");

    let join_handle = tokio::spawn(async move {
        loop {
            let event = event_rx.recv().await;

            match event {
                Some(event) => {
                    log::debug!("Got event: {:?}", &event);

                    for rule in &config.rules {
                        if let Err(err) = rule.evaluate(&conn, &event).await {
                            log::error!("Got error evaluating rule: {:?} {:?}", rule, err);
                        };
                    }

                    if let Event::DaemonStopping = event {
                        log::info!("Stoping event handler task");

                        break;
                    }
                }
                None => {
                    log::warn!("Got an empty event");
                }
            }
        }

        Ok(())
    });

    Ok(join_handle)
}

#[derive(Serialize, Deserialize, Clone, Debug)]
#[serde(tag = "type")]
pub enum Trigger {
    ConnectivityStateChanged {
        states: HashSet<ConnectivityState>,
    },
    DaemonStarted,
    DaemonStopping,
    DeviceAdded {
        device_identifier: Option<DeviceIdentifier>,
    },
    DeviceRemoved {
        device_identifier: Option<DeviceIdentifier>,
    },
    DeviceActivated {
        device_identifier: Option<DeviceIdentifier>,
    },
    DeviceDeactivating {
        device_identifier: Option<DeviceIdentifier>,
    },
    DeviceDisconnected {
        device_identifier: Option<DeviceIdentifier>,
    },
    StateChanged {
        states: HashSet<State>,
    },
}

impl Trigger {
    pub async fn matches_event(&self, conn: &Arc<SyncConnection>, event: &Event) -> Result<bool> {
        match (event, self) {
            (
                Event::ConnectivityStateChanged { connectivity_state },
                Trigger::ConnectivityStateChanged { states },
            ) => Ok(states.contains(connectivity_state)),
            (Event::DeviceAdded { device_path }, Trigger::DeviceAdded { device_identifier }) => {
                match device_identifier {
                    Some(device_identified) => {
                        let trigger_device_path = device_identified.to_path(conn).await?;

                        Ok(device_path.eq(&trigger_device_path))
                    }
                    None => Ok(true),
                }
            }
            (
                Event::DeviceRemoved { device_path },
                Trigger::DeviceRemoved { device_identifier },
            ) => match device_identifier {
                Some(device_identified) => {
                    let trigger_device_path = device_identified.to_path(conn).await?;

                    Ok(device_path.eq(&trigger_device_path))
                }
                None => Ok(true),
            },
            (
                Event::DeviceActivated { device_path },
                Trigger::DeviceActivated { device_identifier },
            ) => match device_identifier {
                Some(device_identified) => {
                    let trigger_device_path = device_identified.to_path(conn).await?;

                    Ok(device_path.eq(&trigger_device_path))
                }
                None => Ok(true),
            },
            (
                Event::DeviceDeactivating { device_path },
                Trigger::DeviceDeactivating { device_identifier },
            ) => match device_identifier {
                Some(device_identified) => {
                    let trigger_device_path = device_identified.to_path(conn).await?;

                    Ok(device_path.eq(&trigger_device_path))
                }
                None => Ok(true),
            },
            (
                Event::DeviceDisconnected { device_path },
                Trigger::DeviceDisconnected { device_identifier },
            ) => match device_identifier {
                Some(device_identified) => {
                    let trigger_device_path = device_identified.to_path(conn).await?;

                    Ok(device_path.eq(&trigger_device_path))
                }
                None => Ok(true),
            },
            (Event::StateChanged { state }, Trigger::StateChanged { states }) => {
                Ok(states.contains(state))
            }
            (Event::DaemonStarted, Trigger::DaemonStarted)
            | (Event::DaemonStopping, Trigger::DaemonStopping) => Ok(true),
            _ => Ok(false),
        }
    }
}
